#!/usr/bin/env python3
"""
DNS Analyzer
Advanced DNS analysis and security assessment tool.
Parameters:
    - domain: Target domain to analyze
    - mode: Analysis mode (basic, full, security)
    - output: Output file for results
Requires: dnspython package
"""
import dns.resolver
import dns.zone
import dns.query
import dns.name
import argparse
import logging
from datetime import datetime
import json
import threading
import queue
from collections import defaultdict
import socket
import whois
import requests
import concurrent.futures
import time

class DNSAnalyzer:
    def __init__(self, domain, mode='full'):
        self.domain = domain
        self.mode = mode.lower()
        self.results = defaultdict(list)
        self.setup_logging()
        self.initialize_resolvers()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def initialize_resolvers(self):
        """Initialize DNS resolvers with different providers"""
        self.resolvers = {
            'google': dns.resolver.Resolver(configure=False),
            'cloudflare': dns.resolver.Resolver(configure=False),
            'quad9': dns.resolver.Resolver(configure=False)
        }
        
        # Set nameservers for each resolver
        self.resolvers['google'].nameservers = ['8.8.8.8', '8.8.4.4']
        self.resolvers['cloudflare'].nameservers = ['1.1.1.1', '1.0.0.1']
        self.resolvers['quad9'].nameservers = ['9.9.9.9', '149.112.112.112']

    def query_record(self, record_type, resolver_name='google'):
        """Query specific DNS record type"""
        try:
            resolver = self.resolvers[resolver_name]
            answers = resolver.resolve(self.domain, record_type)
            return [str(rdata) for rdata in answers]
        except dns.resolver.NXDOMAIN:
            return []
        except dns.resolver.NoAnswer:
            return []
        except Exception as e:
            self.logger.error(
                f"Error querying {record_type} records: {str(e)}"
            )
            return []

    def analyze_basic_records(self):
        """Analyze basic DNS records"""
        record_types = ['A', 'AAAA', 'MX', 'NS', 'TXT', 'SOA', 'CNAME']
        
        for record_type in record_types:
            records = self.query_record(record_type)
            if records:
                self.results['basic_records'].append({
                    'type': record_type,
                    'records': records
                })

    def analyze_security_records(self):
        """Analyze security-related DNS records"""
        # Check SPF records
        txt_records = self.query_record('TXT')
        spf_records = [r for r in txt_records if r.startswith('v=spf1')]
        if spf_records:
            self.results['security_records'].append({
                'type': 'SPF',
                'records': spf_records,
                'analysis': self.analyze_spf(spf_records[0])
            })

        # Check DMARC records
        try:
            dmarc_records = self.query_record('TXT', f'_dmarc.{self.domain}')
            if dmarc_records:
                self.results['security_records'].append({
                    'type': 'DMARC',
                    'records': dmarc_records,
                    'analysis': self.analyze_dmarc(dmarc_records[0])
                })
        except:
            pass

        # Check DNSSEC
        try:
            ds_records = self.query_record('DS')
            dnskey_records = self.query_record('DNSKEY')
            self.results['security_records'].append({
                'type': 'DNSSEC',
                'enabled': bool(ds_records and dnskey_records),
                'ds_records': ds_records,
                'dnskey_records': dnskey_records
            })
        except:
            self.results['security_records'].append({
                'type': 'DNSSEC',
                'enabled': False
            })

    def analyze_spf(self, spf_record):
        """Analyze SPF record configuration"""
        analysis = {
            'valid': True,
            'includes': [],
            'ip_ranges': [],
            'all_mechanism': 'neutral',
            'issues': []
        }

        mechanisms = spf_record.split()
        
        for mechanism in mechanisms:
            if mechanism.startswith('include:'):
                analysis['includes'].append(mechanism[8:])
            elif mechanism.startswith('ip4:') or mechanism.startswith('ip6:'):
                analysis['ip_ranges'].append(mechanism)
            elif mechanism in ['-all', '~all', '?all', '+all']:
                analysis['all_mechanism'] = {
                    '-all': 'fail',
                    '~all': 'softfail',
                    '?all': 'neutral',
                    '+all': 'pass'
                }.get(mechanism, 'neutral')

        # Check for issues
        if analysis['all_mechanism'] == 'pass':
            analysis['issues'].append(
                'SPF record uses +all which is too permissive'
            )
        if len(analysis['includes']) > 10:
            analysis['issues'].append(
                'Too many DNS lookups in SPF record (max 10 recommended)'
            )

        return analysis

    def analyze_dmarc(self, dmarc_record):
        """Analyze DMARC record configuration"""
        analysis = {
            'valid': True,
            'policy': None,
            'subdomain_policy': None,
            'reporting': {},
            'issues': []
        }

        parts = dmarc_record.split(';')
        for part in parts:
            part = part.strip()
            if '=' in part:
                key, value = part.split('=', 1)
                key = key.strip()
                value = value.strip()

                if key == 'p':
                    analysis['policy'] = value
                elif key == 'sp':
                    analysis['subdomain_policy'] = value
                elif key == 'rua':
                    analysis['reporting']['aggregate'] = value
                elif key == 'ruf':
                    analysis['reporting']['forensic'] = value

        # Check for issues
        if not analysis['policy']:
            analysis['issues'].append('Missing required policy (p) tag')
        elif analysis['policy'] == 'none':
            analysis['issues'].append(
                'DMARC policy is set to none, providing no protection'
            )
        
        if not analysis['reporting'].get('aggregate'):
            analysis['issues'].append(
                'No aggregate reporting configured'
            )

        return analysis

    def analyze_zone_transfer(self):
        """Attempt zone transfer and analyze results"""
        ns_records = self.query_record('NS')
        zone_transfers = []

        for ns in ns_records:
            try:
                zone = dns.zone.from_xfr(
                    dns.query.xfr(ns, self.domain)
                )
                records = []
                for name, node in zone.nodes.items():
                    rdatasets = node.rdatasets
                    for rdataset in rdatasets:
                        records.extend([str(rdata) for rdata in rdataset])
                
                zone_transfers.append({
                    'nameserver': ns,
                    'success': True,
                    'records': records
                })
            except:
                zone_transfers.append({
                    'nameserver': ns,
                    'success': False
                })

        self.results['zone_transfer'] = zone_transfers

    def analyze_subdomains(self):
        """Analyze subdomains and their configurations"""
        common_subdomains = [
            'www', 'mail', 'ftp', 'smtp', 'pop', 'ns1', 'webmail',
            'login', 'dev', 'staging', 'api', 'admin', 'test'
        ]

        with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
            future_to_subdomain = {
                executor.submit(
                    self.analyze_subdomain, f"{sub}.{self.domain}"
                ): sub for sub in common_subdomains
            }

            for future in concurrent.futures.as_completed(future_to_subdomain):
                subdomain = future_to_subdomain[future]
                try:
                    result = future.result()
                    if result:
                        self.results['subdomains'].append(result)
                except Exception as e:
                    self.logger.error(
                        f"Error analyzing subdomain {subdomain}: {str(e)}"
                    )

    def analyze_subdomain(self, subdomain):
        """Analyze a specific subdomain"""
        try:
            ip_addresses = self.query_record('A', subdomain)
            if ip_addresses:
                return {
                    'name': subdomain,
                    'ip_addresses': ip_addresses,
                    'records': {
                        'mx': self.query_record('MX', subdomain),
                        'txt': self.query_record('TXT', subdomain)
                    }
                }
        except:
            return None

    def analyze_reverse_dns(self):
        """Perform reverse DNS analysis"""
        a_records = self.query_record('A')
        reverse_records = []

        for ip in a_records:
            try:
                hostname = socket.gethostbyaddr(ip)[0]
                reverse_records.append({
                    'ip': ip,
                    'hostname': hostname,
                    'match': hostname.endswith(self.domain)
                })
            except:
                reverse_records.append({
                    'ip': ip,
                    'hostname': None,
                    'match': False
                })

        self.results['reverse_dns'] = reverse_records

    def analyze_whois(self):
        """Analyze WHOIS information"""
        try:
            w = whois.whois(self.domain)
            self.results['whois'] = {
                'registrar': w.registrar,
                'creation_date': str(w.creation_date),
                'expiration_date': str(w.expiration_date),
                'last_updated': str(w.updated_date),
                'status': w.status,
                'nameservers': w.name_servers
            }
        except Exception as e:
            self.logger.error(f"Error getting WHOIS info: {str(e)}")

    def save_results(self, output_file):
        """Save analysis results to file"""
        scan_results = {
            'timestamp': datetime.now().isoformat(),
            'domain': self.domain,
            'mode': self.mode,
            'results': dict(self.results)
        }

        try:
            with open(output_file, 'w') as f:
                json.dump(scan_results, f, indent=4)
            self.logger.info(f"Results saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving results: {str(e)}")

    def print_summary(self):
        """Print analysis summary"""
        self.logger.info("\nDNS Analysis Summary:")
        self.logger.info(f"Domain: {self.domain}")
        
        # Basic records summary
        if self.results['basic_records']:
            self.logger.info("\nBasic DNS Records:")
            for record in self.results['basic_records']:
                self.logger.info(
                    f"{record['type']}: {', '.join(record['records'])}"
                )

        # Security records summary
        if self.results['security_records']:
            self.logger.info("\nSecurity Records:")
            for record in self.results['security_records']:
                if record['type'] in ['SPF', 'DMARC']:
                    self.logger.info(
                        f"\n{record['type']}:"
                        f"\n  Records: {', '.join(record['records'])}"
                        f"\n  Issues: {', '.join(record['analysis']['issues'])}"
                    )
                elif record['type'] == 'DNSSEC':
                    self.logger.info(
                        f"\nDNSSEC: {'Enabled' if record['enabled'] else 'Disabled'}"
                    )

        # Zone transfer results
        if self.results['zone_transfer']:
            self.logger.info("\nZone Transfer Results:")
            for zt in self.results['zone_transfer']:
                self.logger.info(
                    f"Nameserver {zt['nameserver']}: "
                    f"{'Vulnerable' if zt['success'] else 'Protected'}"
                )

        # Subdomain summary
        if self.results['subdomains']:
            self.logger.info(f"\nSubdomains Found: {len(self.results['subdomains'])}")
            for subdomain in self.results['subdomains']:
                self.logger.info(
                    f"\n{subdomain['name']}:"
                    f"\n  IP: {', '.join(subdomain['ip_addresses'])}"
                )

        # WHOIS summary
        if 'whois' in self.results:
            whois_info = self.results['whois']
            self.logger.info(
                f"\nWHOIS Information:"
                f"\n  Registrar: {whois_info['registrar']}"
                f"\n  Creation Date: {whois_info['creation_date']}"
                f"\n  Expiration Date: {whois_info['expiration_date']}"
            )

    def start_analysis(self, output_file=None):
        """Start DNS analysis"""
        self.logger.info(f"Starting DNS analysis for {self.domain}")
        self.logger.info(f"Analysis mode: {self.mode}")
        
        try:
            # Basic analysis
            self.analyze_basic_records()
            
            # Security analysis
            if self.mode in ['full', 'security']:
                self.analyze_security_records()
                self.analyze_zone_transfer()
            
            # Full analysis
            if self.mode == 'full':
                self.analyze_subdomains()
                self.analyze_reverse_dns()
                self.analyze_whois()
                
        except KeyboardInterrupt:
            self.logger.info("Analysis interrupted by user")
        except Exception as e:
            self.logger.error(f"Error during analysis: {str(e)}")
        finally:
            if output_file:
                self.save_results(output_file)
            self.print_summary()

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="DNS Analyzer")
    parser.add_argument("domain", help="Target domain to analyze")
    parser.add_argument("--mode", choices=['basic', 'full', 'security'],
                      default='full', help="Analysis mode")
    parser.add_argument("--output", help="Output file for analysis results")
    args = parser.parse_args()

    analyzer = DNSAnalyzer(args.domain, args.mode)
    analyzer.start_analysis(args.output)
